"""
um80 - Microsoft MACRO-80 compatible assembler toolchain for Linux.

This package provides Unix/Linux implementations of the classic CP/M
development tools from Microsoft:

- um80: MACRO-80 compatible assembler
- ul80: LINK-80 compatible linker
- ulib80: LIB-80 compatible library manager
- ucref80: Cross-reference utility
- ud80: 8080/Z80 disassembler
- ux80: 8080 to Z80 assembly translator
"""

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("um80")
except PackageNotFoundError:
    __version__ = "unknown"
